#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# **********************************************************************************************************************
# File: config.py
# Project: ORCA
# Created: April 11, 2023, 12:25h
# Author: Zacharias Chalampalakis 
# Email: Zacharias.Chalampalakis@meduniwien.ac.at
# Institute: Quantitative Imaging and Medical Physics, Medical University of Vienna
# Description: This module contains the constants for the wolfz.
# License: Apache 2.0
# **********************************************************************************************************************

import os 
from orcaz import constants
from orcaz.file_utilities import create_directory

orca_dir = ""
synth_dir = ""
